/*
 */

#include "sm/generic.h"
SM_RCSID("@(#)$Id: sm-conf-type-ipv6.c,v 1.5 2013/09/22 20:23:09 ca Exp $")

#if SM_LIBCONF_ALONE
#include <limits.h>
#include <errno.h>
#include <ctype.h>
#include <string.h>
#include "sm-conf.h"
#include "sm-util.h"
#else /* SM_LIBCONF_ALONE */
#include "sm/limits.h"
#include "sm/error.h"
#include "sm/ctype.h"
#include "sm/memops.h"
#include "sm/string.h"
#include "sm/sm-conf.h"
#include "sm/net.h"
#endif /* SM_LIBCONF_ALONE */

#include "sm-conf-node.h"
#include "sm-conf-state.h"
#include "sm-conf-type.h"
#include "sm-conf-util.h"

#if MTA_NETINET6
/*
**  SM_MEMTOIPV6 --
**
**	Parameters:
**		text -- the text to be converted
**		text_n -- # of bytes pointed to by text
**		endptr -- if non-NULL, a pointer to the first
**			unconverted byte is stored here.
**		val_out -- value is stored here.
**
**	Returns:
**		EINVAL for a format error
**		ERANGE for overflow
**		0 for success.
*/

static sm_ret_T
sm_memtoipv6(
	char const		*text,
	size_t			text_n,
	char			**endptr,
	ipv6_P			val_out)
{
	size_t n;
	sm_ret_T ret;
	ipv6_T ipv6;
	char ipv6str[INET6_ADDRSTRLEN];

	if (text_n >= sizeof(ipv6str))
		n = sizeof(ipv6str) - 1;
	else
		n = text_n;
	sm_memcpy(ipv6str, text, n);
	ipv6str[n] = '\0';
	ret = sm_inet_a2ipv6(ipv6str, endptr, &ipv6);
	sm_memcpy(val_out, &ipv6, sizeof(ipv6_T));
	return ret;
}

/* BROKEN XXX */
static int
ipv6_write(
	sm_conf_definition_T const	*def,
	void			*data,
	ipv6_P			ipv6p)
{
	SM_IS_CONF_DEF(def);
	if (def->scd_size != sizeof(ipv6_T))
		return SM_CONF_ERR_INVALID;

	sm_memcpy(data, ipv6p, sizeof(ipv6_T));
	return 0;
}

static int
ipv6_scan(
	sm_conf_T		*smc,
	sm_conf_node_T const	*node,
	sm_conf_definition_T const *def,
	char const		*text,
	size_t			text_n,
	ipv6_P			val_out)
{
	char		loc[SM_CONF_ERROR_BUFFER_SIZE];
	size_t		i;
	int			err;
	char		*endptr;

	SM_IS_CONF_DEF(def);

	err = sm_memtoipv6(text, text_n, &endptr, val_out);
	if (err != 0) {
		sm_conf_error_add(smc, "%s: expected IPv6 address, got '%.*s'",
			sm_conf_node_location(smc, node, loc, sizeof loc),
			(int)text_n, text);
		return SM_CONF_ERR_TYPE;
	}

	i = endptr - text;

	/* If there's something left, and we have suffixes, use them. */
	if (i < text_n) {
		sm_conf_error_add(smc, "%s: unexpected trailing "
			"characters in '%.*s'",
			sm_conf_node_location(smc, node, loc, sizeof loc),
			(int)text_n, text);
		return SM_CONF_ERR_TYPE;
	}

	return 0;
}

static int
sm_conf_type_ipv6_node_to_value(
	sm_conf_T			*smc,
	sm_conf_definition_T const	*def,
	uint32_t			 rqmnts,
	sm_conf_node_T		*node,
	void				*data)
{
	char const		*text;
	size_t			text_n;
	int				err;
	ipv6_T			ipv6;

	SM_IS_CONF_DEF(def);
	err = sm_conf_node_to_value(smc, "ipv6ean", node, &text, &text_n);
	if (err != 0) return err;

	err = ipv6_scan(smc, node, def, text, text_n, &ipv6);
	if (err != 0) return err;

	if (data != NULL)
		err = ipv6_write(def, data, &ipv6);

	return err;
}

static int
sm_conf_type_ipv6_value_check(
	sm_conf_T			*smc,
	sm_conf_definition_T const	*def,
	void const			*data)
{
	SM_IS_CONF_DEF(def);

	/* if we have a check function, use it. */
	if (def->scd_check != NULL)
		return (* def->scd_check)(smc, def->scd_check_data, def, data);
	return 0;
}

static int
sm_conf_type_ipv6_value_null(
	sm_conf_T			*smc,
	sm_conf_definition_T const	*def,
	void				*data)
{
	ipv6_T			ipv6;
	int				err;

	if (data != NULL) {
		SM_IS_CONF_DEF(def);
		sm_memzero(&ipv6, sizeof(ipv6));

		/* is there a default? if NULL: don't set value to 0 */
		if (def->scd_default != NULL) {
			/* is there an actual value? */
			if (*def->scd_default != '\0') {
				err = ipv6_scan(smc, NULL, def, def->scd_default,
					strlen(def->scd_default), &ipv6);
				if (err != 0) return err;
			}
			err = ipv6_write(def, data, &ipv6);
			if (err != 0)
				return err;
		}
	}
	return 0;
}

sm_conf_type_T const
sm_conf_type_ipv6_data =
{
	sm_conf_type_ipv6_node_to_value,
	sm_conf_type_ipv6_value_check,
	sm_conf_type_ipv6_value_null
};
#endif /* MTA_NETINET6 */
