/*
 *  Copyright (c) 2005, 2007, 2008 Sendmail, Inc. and its suppliers.
 *    All rights reserved.
 *
 * By using this file, you agree to the terms and conditions set
 * forth in the LICENSE file which can be found at the top level of
 * the sendmail distribution.
 */

#include "sm/generic.h"
SM_RCSID("@(#)$Id: mail_parse.c,v 1.6 2011/01/30 21:49:06 ca Exp $")
#include "sm/limits.h"
#include "sm/assert.h"
#include "sm/error.h"
#include "sm/memops.h"
#include "sm/ctype.h"
#include "sm/mail_parse.h"

/* types */
typedef unsigned long cmap_elem_type;

/* XREF: error.h! (not used outside this module for now...) */
#define R2822_BASE	(R2821_BASE + 64)

/* symbolic names */
#define RFC2822_OK			SM_SUCCESS
/* unbalanced parentheses */
#define RFC2822_ERR_PUNBALANCED		sm_error_perm(SM_EM_ADDR, R2822_BASE+0)
/* unbalanced quotes */
#define RFC2822_ERR_QUNBALANCED		sm_error_perm(SM_EM_ADDR, R2822_BASE+1)
/* unbalanced sq. brackets */
#define RFC2822_ERR_SUNBALANCED		sm_error_perm(SM_EM_ADDR, R2822_BASE+2)
#define RFC2822_ERR_UNBALANCED		sm_error_perm(SM_EM_ADDR, R2822_BASE+3)

/* a bitmap for the "specials" character class */
#define	CMAP_NBITS		(sizeof(cmap_elem_type) * CHAR_BIT)
#define	CMAP_NELEMS		((1 + UCHAR_MAX) / CMAP_NBITS)
#define	CMAP_INDEX(i)	((i) / CMAP_NBITS)
#define	CMAP_BIT(i)		(1L << (i) % CMAP_NBITS)
#define	CMAP_TST(ar, c)	((ar)[CMAP_INDEX(c)] &  CMAP_BIT(c))
#define	CMAP_SET(ar, c)	((ar)[CMAP_INDEX(c)] |= CMAP_BIT(c))

static uchar const SPECIALS[] = "<>@,;:\\\"/[]?=";
static cmap_elem_type is_special[CMAP_NELEMS] = { 0 };

/*
**  RFC2822_INIT -- set up data for address parsing
**
**  Parameters:
**		none
**
**  Return value:
**		none
*/

void
rfc2822_init(void)
{
	uint u;

	/* set up special finder */
	for (u = 0; SPECIALS[u] != '\0'; u++)
		CMAP_SET(is_special, SPECIALS[u]);
}

/*
**  MATCHING_PAREN -- return the location past matching opposite parentheses
**
**  Parameters:
**		s -- start of string to be processed
**		e -- end of string to be processed
**		open_paren -- open parenthesis character
**		close_paren -- close parenthesis character
**
**  Return value:
**		Location of the final close parenthesis character in the string.
**		For example, given "xxx((yyyy)zz)aaaa", would return the location
**		of the second ")".  There may be more beyond that, but at that point
**		everything is balanced.
*/

static sm_ret_T
matching_paren(uchar *s, const uchar *e, uchar open_paren, uchar close_paren, uchar **closing)
{
	int	paren;

	paren = 1;
	SM_REQUIRE(s != NULL);
	SM_REQUIRE(e != NULL);
	for (; s < e; s++) {
		if (*s == close_paren) {
			if (--paren == 0) {
				*closing = s;
				return SM_SUCCESS;
			}
		}
		else if (*s == open_paren)
			paren++;
		else if (*s == '\\') {
			if (s + 1 < e && s[1] != '\0')
				s++;
		}
	}
	return RFC2822_ERR_UNBALANCED;
}

/*
**  RFC2822_FIRST_SPECIAL -- find the first RFC2822 "special" character
**
**  Parameters:
**		p -- input string
**		e -- end of input string
**		special_out -- pointer to the first special character found
**
**  Return value:
**		0 on success, or an RFC2822_ERR_* on failure.
*/

static sm_ret_T
rfc2822_first_special(uchar *p, uchar *e, uchar **special_out)
{
	sm_ret_T ret;
	uchar *at_ptr;

	SM_REQUIRE(p != NULL);
	SM_REQUIRE(e != NULL);
	SM_REQUIRE(special_out != NULL);
	at_ptr = NULL;
	for (; p < e && *p != '\0'; p++) {
		/* skip white space between tokens */
		while (p < e && (*p == '(' || (ISASCII(*p) && ISSPACE(*p)))) {
			if (*p != '(')
				p++;
			else {
				ret = matching_paren(p + 1, e, '(', ')', &p);
				if (sm_is_err(ret))
					return RFC2822_ERR_PUNBALANCED;
				p++;
			}
		}

		if (*p == '\0')
			break;

		if (*p == '"') {
			ret = matching_paren(p + 1, e, '\0', '"', &p);
			if (sm_is_err(ret))
				return RFC2822_ERR_QUNBALANCED;
		}
		else if (*p == '[') {
			ret = matching_paren(p + 1, e, '\0', ']', &p);
			if (sm_is_err(ret))
				return RFC2822_ERR_SUNBALANCED;
		}
		else if (CMAP_TST(is_special, *p)) {
			if (*p == '<') {
				*special_out = p;
				return 0;
			}
			else if (*p == ':' || *p == ';' || *p == ',') {
				if (at_ptr != NULL)
					*special_out = at_ptr;
				else
					*special_out = p;
				return 0;
			}
			else if (*p == '@')
				at_ptr = p;
		}
		else {
			while (*p != '\0' && !CMAP_TST(is_special, *p) &&
			       (!ISASCII(*p) || !ISSPACE(*p)) && *p != '(')
				p++;
			p--;
		}
	}

	*special_out = p;
	return 0;
}

/*
**  RFC2822_TOKEN -- find the next token
**
**  Parameters:
**		s -- start of input string
**		e -- end of input string
**		type_out -- type of token (returned)
**		start_out -- start of token (returned)
**		end_out -- end of token (returned)
**		uncommented_whitespace -- set if uncommented whitespace is
**	                          discovered (returned)
**
**  Return value:
**		0 on success, or an RFC2822_ERR_* on failure.
*/

static sm_ret_T
rfc2822_token(uchar *s, uchar *e, int *type_out, uchar **start_out,
              uchar **end_out, bool *uncommented_ws)
{
	sm_ret_T ret;
	int token_type;
	uchar *p, *token_start, *token_end;

	*start_out = NULL;
	*end_out   = NULL;
	*type_out  = 0;
	ret = SM_SUCCESS;

	p = s;

	/* skip white space between tokens */
	while (p < e && (*p == '(' || (ISASCII(*p) && ISSPACE(*p)))) {
		if (*p != '(') {
			*uncommented_ws = true;
			p++;
		}
		else {
			ret = matching_paren(p + 1, e, '(', ')', &p);
			if (sm_is_err(ret))
				return RFC2822_ERR_PUNBALANCED;
			p++;
		}
	}

	if (p >= e || *p == '\0')
		return SM_SUCCESS;

	/* our new token starts here */
	token_start = p;

	/* fill in the token contents and type */
	if (*p == '"') {
		ret = matching_paren(p + 1, e, '\0', '"', &token_end);
		if (sm_is_err(ret))
			return RFC2822_ERR_QUNBALANCED;
		token_type = '"';
		token_end++;
	}
	else if (*p == '[') {
		ret = matching_paren(p + 1, e, '\0', ']', &token_end);
		if (sm_is_err(ret))
			return RFC2822_ERR_SUNBALANCED;
		token_type = '[';
		token_end++;
	}
	else if (CMAP_TST(is_special, *p)) {
		token_end  = p + 1;
		token_type = *p;
	}
	else {
		while (p < e && *p != '\0' && !CMAP_TST(is_special, *p) &&
		       (!ISASCII(*p) || !ISSPACE(*p)) && *p != '(')
			p++;

		token_end = p;
		token_type = 'x';
	}

	*start_out = token_start;
	*end_out   = token_end;
	*type_out  = token_type;

	return ret;
}

/*
**  RFC2822_MAILBOX_SPLIT -- extract the local-part and hostname from an
**                           RFC2822 header, e.g. "From:"
**
**  Parameters:
**		line -- input line
**		user_out -- pointer to "local-part" (returned)
**		domain_out -- pointer to hostname (returned)
**
**  Return value:
**		0 on success, or an RFC2822_ERR_* on failure.
**
**  Notes:
**		Input string is modified.
*/

int
rfc2822_mailbox_split(uchar *line, uchar **user_out, uchar **domain_out)
{
	int type, err;
	bool ws;
	uchar *e, *special;
	uchar *tok_s, *tok_e;
	uchar *w;

	*user_out = NULL;
	*domain_out = NULL;

	err = 0;
	w = line;
	e = line + strlen((const char *)line);
	ws = false;

	for (;;) {
		err = rfc2822_first_special(line, e, &special);
		if (err != 0)
			return err;

		/* given the construct we're looking at, do the right thing */
		switch (*special) {
		  case '<':
			/* display name <address> */
			line = special + 1;
			for (;;) {
				err = rfc2822_token(line, e, &type, &tok_s, &tok_e, &ws);
				if (sm_is_err(err))
					return err;

				if (type == '>' || type == '\0') {
					*w = '\0';
					return SM_SUCCESS;
				}
				else if (type == '@') {
					*w++ = '\0';
					*domain_out = w;
				}
				else if (type == ',' || type == ':') {
					/* source route punctuation */
					*user_out = NULL;
					*domain_out = NULL;
				}
				else {
					if (*user_out == NULL)
						*user_out = w;
					memmove(w, tok_s, tok_e - tok_s);
					w += tok_e - tok_s;
				}
				line = tok_e;
			}
			return SM_SUCCESS;

		  case ';':
		  case ':':
		  case ',':
			/* skip a group name or result */
			line = special + 1;
			break;

		  default:
			/* (display name) addr(display name)ess */
			ws = false;
			for (;;) {
				err = rfc2822_token(line, e, &type, &tok_s, &tok_e, &ws);
				if (sm_is_err(err))
					return err;

				if (type == '\0' ||  type == ',' || type == ';') {
					*w = '\0';
					break;
				}
				else if (type == '@') {
					*w++ = '\0';
					*domain_out = w;
					ws = false;
				}
				else {
					if (*user_out == NULL)
						*user_out = w;
					else if (type == 'x' && ws)
						*w++ = ' ';

					memmove(w, tok_s, tok_e - tok_s);
					w += tok_e - tok_s;

					ws = false;
				}
				line = tok_e;
			}
			return SM_SUCCESS;
		}
	}
	return err;
}
